/*
 * led_prog.c
 *
 *  Created on: Aug 3, 2023
 *      Author: true
 *
 */


#include <stdint.h>

#include "btn.h"
#include "hsv2rgb.h"
#include "led_sk6x_spi.h"
#include "led_user.h"
#include "prng.h"
#include "userconf.h"



#define LEDPROG_MAX_PROGRAMS		7



void ledprog0_twinkle_white();
void ledprog1_twinkle_rgb_rand();
void ledprog2_twinkle_rgb_set();
void ledprog3_rainbow();
void ledprog4_flicker_same();
void ledprog5_flicker_all();
void ledprog6_ramper();



const void (*led_prog_list[])() = {
	ledprog0_twinkle_white,
	ledprog1_twinkle_rgb_rand,
	ledprog2_twinkle_rgb_set,
	ledprog3_rainbow,
	ledprog4_flicker_same,
	ledprog5_flicker_all,
	ledprog6_ramper
};

static uint8_t  led_prog_active = 0;
static uint32_t work[4];

static uint8_t r, g, b;

static uint8_t prog_mode = 0;
static uint8_t prog_tmr = 0;
static uint8_t prog_led = 0;
static uint8_t prog_idx = 0;

/*
 * program pages:
 * 1-4: 4 program bytes for active led program
 */



void ledprog_default()
{
	uconf.led[3][2] = 2;
	uconf.led[3][3] = 30;

	uconf.led[4][1] = 248;
	uconf.led[4][2] = 170;
	uconf.led[4][3] = 16;

	uconf.led[6][1] = 192;
	uconf.led[6][2] = 30;
	uconf.led[6][3] = 6;
}

void ledprog_change(uint8_t idx)
{
	uint8_t i;

	led_prog_active = idx;

	for (i = 0; i < 4; i++) {
		work[i] = 0;
	}
}

void ledprog_btn_push_cb(uint8_t idx)
{
	if (!prog_mode) return;

	switch (idx) {
		case BTN_MODE: {

		}
		case BTN_SET: {

		}
	}
}

void ledprog_btn_held_cb(uint8_t idx)
{
	if (idx == BTN_PROG) {
		prog_mode ^= 1;
		prog_idx = prog_tmr = 0;

		// when leaving programming mode we need to commit settings to flash
		if (!prog_mode) {
			uconf.led_prog_idx = led_prog_active;
			uconf_save();
		}

		return;
	}

	if (prog_mode) {
		switch (idx) {
			case BTN_MODE:
			case BTN_SET: {
				ledprog_btn_push_cb(idx);
				break;
			}
		}

		// allow retrigger
		btn[idx].state &= ~(BTN_HELDRT | BTN_HELDRT_CB);
	} else {
		// do not retrigger on these
		if (btn[idx].state & BTN_HELDRT_CB) return;

		switch (idx) {
			case BTN_MODE: {	// return to user's default program
				ledprog_change(uconf.led_prog_idx);
				led_sk6x_brightness(uconf.led_bright ? uconf.led_bright : 1);
				break;
			}
			case BTN_SET: {		// change active zones
				uconf.led_zone++;
				if (uconf.led_zone > 7) uconf.led_zone = 1;
				led_sk6x_zone_ena(uconf.led_zone);
				break;
			}
		}
	}
}

void ledprog_btn_release_cb(uint8_t idx)
{
	if (btn[idx].state & BTN_HELD) {
		btn[idx].state = 0;
		return;
	}

	if (prog_mode) {
		switch (idx) {
			case BTN_PROG: {	// change program variable page
				prog_idx++;
				if (prog_idx >= 4) prog_idx = 0;
				break;
			}
		}
	} else {
		switch (idx) {
			case BTN_MODE: {	// change active program
				led_prog_active++;
				if (led_prog_active >= LEDPROG_MAX_PROGRAMS) led_prog_active = 0;
				ledprog_change(led_prog_active);

				// increase brightness if we are dark
				if (uconf.led_bright) {
					break;
				}
			}
			case BTN_PROG: {	// change brightness
				uconf.led_bright++;
				if (uconf.led_bright > 5) uconf.led_bright = 0;
				led_sk6x_brightness(uconf.led_bright);
				break;
			}
			case BTN_SET: {		// do nothing?

			}
		}
	}

	// clear any button state as button is now idle
	btn[idx].state = 0;
}

void ledprog_btn()
{
	btn[BTN_PROG].held_cb     = ledprog_btn_held_cb;
	btn[BTN_PROG].release_cb  = ledprog_btn_release_cb;
	btn[BTN_PROG].hold_thresh = BTN_HOLD_1_00S;

	btn[BTN_MODE].push_cb     = ledprog_btn_push_cb;
	btn[BTN_MODE].held_cb     = ledprog_btn_held_cb;
	btn[BTN_MODE].release_cb  = ledprog_btn_release_cb;
	btn[BTN_MODE].hold_thresh = BTN_HOLD_1_00S;
	btn[BTN_MODE].hold_retrig = 10;

	btn[BTN_SET].push_cb      = ledprog_btn_push_cb;
	btn[BTN_SET].held_cb      = ledprog_btn_held_cb;
	btn[BTN_SET].release_cb   = ledprog_btn_release_cb;
	btn[BTN_SET].hold_thresh  = BTN_HOLD_1_00S;
	btn[BTN_SET].hold_retrig  = 10;
}

void ledprog_run()
{
	if (prog_mode) {
		// we need to see the LEDs
		if (!uconf.led_bright) {
			uconf.led_bright++;
			led_sk6x_brightness(uconf.led_bright);
		}

		// 2 second LED flash timer
		if (!prog_tmr) prog_led = prog_idx + 1;

		// flash LED to indicate the page we are programming
		if (prog_led) {
			switch (prog_tmr & 0x1f) {
				case 0x00: {
					userled_set(255);
					break;
				}
				case 0x0a: {
					userled_set(0);
					prog_led--;
					break;
				}
			}
		}

		prog_tmr++;
	}

	// always run a program
	led_prog_list[led_prog_active]();
}

void ledprogi_twinkle(uint8_t r, uint8_t g, uint8_t b)
{
	uint8_t x;

	x = prng_get8();

	led_sk6x_set_all(0, 0, 0);

	if (x > 240) {
		x = prng_get32() & 0x7;
		led_sk6x_set(x, r, g, b);
	}
}

/*
 * uconf.led:
 * 0	dwell
 * 1	threshold
 */
void ledprog0_twinkle_white()
{
	ledprogi_twinkle(0xff, 0xff, 0xff);
}

/*
 * uconf.led:
 * 0	dwell
 * 1	threshold
 */
void ledprog1_twinkle_rgb_rand()
{
	uint32_t x;

	x = prng_get32();

	// set values
	r = (x >> 24);
	g = (x >> 16) & 0xff;
	b = (x >>  8) & 0xff;

	// randomly reduce colors to remove "whiteness"
	switch (x & 0x3) {
		case 1: {
			r >>= 4;
			g >>= 4;
			break;
		}
		case 2: {
			g >>= 4;
			b >>= 4;
			break;
		}
		case 3: {
			b >>= 4;
			r >>= 4;
			break;
		}
	}

	ledprogi_twinkle(r, g, b);
}


/*
 * uconf.led:
 * 0	dwell
 * 1	threshold
 * 2	hue
 * 3	val
 */
void ledprog2_twinkle_rgb_set()
{
	hsv2rgb_8b(uconf.led[2][2] * 6, 255, uconf.led[2][3], &r, &g, &b);

	ledprogi_twinkle(r, g, b);
}

/*
 * uconf.led:
 * 0	dwell
 * 1	threshold
 * 2	speed
 * 3	skip
 *
 * work:
 * 0	base hue
 */
void ledprog3_rainbow()
{
	uint8_t i;
	uint16_t w;

	work[0] += (uconf.led[3][2] * 6) + 1;
	work[0] %= 1536;
	w = work[0];

	for (i = 0; i < 8; i++) {
		hsv2rgb_8b(w, 255, 255, &r, &g, &b);
		led_sk6x_set(i, r, g, b);

		w += (uconf.led[3][3] * 6) + 1;
		w %= 1536;
	}
}

void ledprogi_flicker()
{
	uint16_t h;
	uint8_t x;

	h = uconf.led[4][2] * 6;

	x = prng_get8();

	if (x > uconf.led[4][1]) {
		x = 255;
	} else {
		x = uconf.led[4][3];
	}

	x &= ~0x07;
	x |= prng_get32() & 0x07;

	hsv2rgb_8b(h, 255, x, &r, &g, &b);
}

/*
 * uconf.led:
 * 0	dwell
 * 1	threshold
 * 2	hue
 * 3	base
 */
void ledprog4_flicker_same()
{
	ledprogi_flicker();
	led_sk6x_set_all(r, g, b);
}

/*
 * uconf.led (uses ledprog4)
 * 0	dwell
 * 1	threshold
 * 2	hue
 * 3	val
 */
void ledprog5_flicker_all()
{
	uint8_t i;

	for (i = 0; i < 8; i++) {
		ledprogi_flicker();
		led_sk6x_set(i, r, g, b);
	}
}

/*
 * uconf.led
 * 0	dwell
 * 1	speed
 * 2	hue
 * 3	faderate
 *
 * work:
 * 0	ramper state
 */
void ledprog6_ramper()
{
	uint8_t i;

	uint16_t h;
	int16_t spd;

	uint16_t floor = 0;
	uint8_t v[8];
	uint8_t fr;


	h = uconf.led[6][2] * 6;
	fr = uconf.led[6][3] & 0x3f;

	// speed -64 to -1, +1 to +64
	spd = (uconf.led[6][1] >> 1) - 64;
	if (spd >= 0) spd++;
	work[0] += spd;
	work[0] &= 0xfff;

	// unpack fade values
	for (i = 0; i < 4; i++) {
		v[0 + i] = work[2] >> (i * 8);
		v[4 + i] = work[3] >> (i * 8);
	}

	for (i = 0; i < 8; i++) {
		// set to max if within range
		if (work[0] >= floor && work[0] <= floor + 64) {
			v[i] = 255;
		}

		// set LED
		hsv2rgb_8b(h, 255, v[i], &r, &g, &b);
		led_sk6x_set(i, r, g, b);

		// set fadeout value
		if (fr > v[i]) {
			v[i] = 0;
		} else {
			v[i] -= fr;
		}

		// both sides of top LEDs match
		if (i == 5) {
			i++;
			led_sk6x_set(i, r, g, b);
		}

		// work up the chain
		floor += 0x180;
	}

	// pack fade values
	work[2] = work[3] = 0;
	for (i = 0; i < 4; i++) {
		work[2] |= v[0 + i] << (i * 8);
		work[3] |= v[4 + i] << (i * 8);
	}
}
